#include <stdio.h>
#include <math.h>
#include "parms.h"
#include "graddata.h"
#include "gradient.h"

static char *sccsversion = "%W%   %G%";

/* replace gradient() body with analytical gradient code, if desired.
   by default, numerical gradients with the intervals in *optint are used

   arguments (must be identical in any replacement code):
   *p, vector of parameters, p[0] is the first parameter;
   *g, vector that will hold the gradient, allocated by caller,
       g[0] is the derivative for the first parameter;
   nparms, the number of parameters, p[nparms-1] is the last parameter.
*/
void gradient(double *p, double *g, int nparms)
{
  extern double (*func4g)(double *);

  static double *wrk = NULL;

  if (wrk==NULL) wrk = (double *) malloc(nparms*sizeof(double));

  /* numgradc:  numerical gradient, central-difference */
  numgradc(nparms, p, g, wrk, func4g);
}

/* numerical gradient, forward-difference */
/* see Gill, Murray and Wright, 1981, "Practical Optimization," p. 342 */

/* invals, grads, wrk should point to double w[nparms+1] arrays;
   func is the function whose gradient is to be evaluated */
void numgrad(int nparms, double *invals, double *grads, double *wrk,
		double (*func)(double *))
{
  int i;
  double u, rf, fplus, fminus;
  double epsuse;
  double duse;

  /* evaluate func at the input point */
  u = func(invals);

  /* copy the parameter values for point at which to evaluate the gradient */
  for (i=0; i<nparms; i++) wrk[i] = invals[i];

  /* evaluate the gradient */
  for (i=0; i<nparms; i++) {
    epsuse = epsacc[i];
    duse = optint[i];
    wrk[i] += duse;
    grads[i] = (func(wrk) - u) / duse ;
    /* check the gradient */
    if (2.0*epsuse / (duse*fabs(grads[i])) > 0.1) { /* switch to central-diff */
      duse = pow(duse, 2.0/3.0);  /* see GMW p 131 */
      wrk[i] = invals[i] + duse;
      fplus = func(wrk);
      wrk[i] = invals[i] - duse;
      fminus = func(wrk);
      grads[i] = (fplus-fminus) * 0.5 / duse ;
    }
    wrk[i] = invals[i];
  }
}

/* numerical gradient, central-difference */
/* see Gill, Murray and Wright, 1981, "Practical Optimization," p. 342 */

/* invals, grads, wrk should point to double w[nparms+1] arrays;
   func is the function whose gradient is to be evaluated */
void numgradc(int nparms, double *invals, double *grads, double *wrk,
		double (*func)(double *))
{
  int i;
  double u, rf, fplus, fminus;
  double epsuse;
  double duse;

  /* evaluate func at the input point */
  u = func(invals);

  /* copy the parameter values for point at which to evaluate the gradient */
  for (i=0; i<nparms; i++) wrk[i] = invals[i];

  /* evaluate the gradient */
  for (i=0; i<nparms; i++) {
    epsuse = epsacc[i];
    duse = optint[i];
    duse = pow(duse, 2.0/3.0);  /* see GMW p 131 */
    wrk[i] = invals[i] + duse;
    fplus = func(wrk);
    wrk[i] = invals[i] - duse;
    fminus = func(wrk);
    grads[i] = (fplus-fminus) * 0.5 / duse ;
    wrk[i] = invals[i];
  }
}

/* numerical outer-product of gradient, central-difference */
/* return pointer to full numerical OPG */
/*  uses central differences: */
/*   phi2 = (f(x+hi ei) - f(x-hi ei)) * (f(x+hj ej) - f(x-hj ej)) */
/* invals, grads, wrk should point to double w[nparms+1] arrays;
   func is the function whose gradient is to be evaluated */
double *numopgc(int nparms, int nobs, double *invals, double *opg, double *wrk,
		int (*func)(double *X, double *outvec))
{
  double *outvec, *outplus, *outminus, *outdiff;

  int i,j,k, idx;
  int ni = 0;
  double hi, ih;
  double phi2, dini;

  double *hfuse = optint;

  outvec = (double *) malloc(nobs*sizeof(double));
  outplus = (double *) malloc(nobs*nparms*sizeof(double));
  outminus = (double *) malloc(nobs*nparms*sizeof(double));
  outdiff = (double *) malloc(nobs*nparms*sizeof(double));

  /* allocate storage for the hessian */
  opg = (double *) calloc(nparms*nparms,sizeof(double));

  /* evaluate func at the input point */
  func(invals-1, outvec);

  /* copy the parameter values for point at which to evaluate the gradient */
  for (i=0; i<nparms; i++) wrk[i] = invals[i];

  /* evaluate the gradient elements */
  for (i=0; i<nparms; i++) {
    hi = pow(hfuse[i], 2.0/3.0);
    wrk[i] = invals[i] + hi;
    func(wrk-1, outplus+i*nobs);
    wrk[i] = invals[i] - hi;
    func(wrk-1, outminus+i*nobs);
    wrk[i] = invals[i];

    ih = 0.5 / hi;
    for (k=0; k<nobs; k++) {
      idx = i*nobs+k;
      outdiff[idx] = (outplus[idx] - outminus[idx]) * ih ;
    }
  }
  for (i=0; i<nparms; i++) {
    idx = i*nparms;
    for (j=0; j<=i; j++) {
      phi2 = 0.0;
      for (k=0; k<nobs; k++) {
	phi2 += outdiff[i*nobs+k] * outdiff[j*nobs+k];
      }
      opg[i*nparms + j] = phi2;
      if (i!=j) opg[j*nparms + i] = phi2;
    }
  }
  free(outdiff);
  free(outminus);
  free(outplus);
  free(outvec);

  return opg;
}

/* estimate accuracy */
/* see Gill, Murray and Wright, 1981, "Practical Optimization," p. 337 */

double **eaccuracy(int nparms, int ndiffs, double h, double *invals,
		   double *wrk, double (*func)(double *))
{
  static double **table = NULL;
  static int tndiffs = 0;

  int i,j,k, idx;
  int nsteps = 1+2*ndiffs, nrows = nparms*nsteps, ncols = ndiffs+1;
  double u, huse, v;
  double scale = 2.0*pow(10.0,6.0);

  /* allocate storage for the table to differences to be returned */
  if (table == NULL || tndiffs != ndiffs) {
    table = (double **) malloc(ncols*sizeof(double *));
    for (i=0; i<ncols; i++)
      table[i] = (double *) calloc(nrows, sizeof(double));
    tndiffs = ndiffs;
  }

  /* evaluate func at the input point */
  u = func(invals);
  for (i=0; i<nparms; i++) table[0][i*nsteps] = u;

  /* copy the parameter values for point at which to evaluate the gradient */
  for (i=0; i<nparms; i++) wrk[i] = invals[i];

  /* evaluate the offsets */
  for (i=0; i<nparms; i++) {
    /* make sure huse is sufficiently small */
    v = fabs(invals[i]);
    huse = h;
    if (v>MACHEPS*scale) {
      while (huse > v/scale)
	huse *= 0.1;
    }
    for (j=1; j<nsteps; j++) {
      wrk[i] += huse;
      table[0][i*nsteps+j] = func(wrk) ;
    }
    wrk[i] = invals[i];
  }

  /* compute the differences */
  for (i=0; i<nparms; i++) {
    idx = i*nsteps;
    for (j=0; j<ndiffs; j++) {
      for (k=0; k<nsteps-j-1; k++) {
	table[j+1][idx+k] = table[j][idx+k+1] - table[j][idx+k];
      }
    }
  }
  return table;
}

/* estimate intervals for use in numerical gradients */
/* see Gill, Murray and Wright, 1981, "Practical Optimization," p. 343-344 */

struct estints *algfd(int nparms, double *eps, double *invals, double *wrk,
		      double (*func)(double *))
{
  static struct estints *outstruc = NULL;

  int i,j,k;
  int K = 20;
  int errval = 0;
  double omega = 1.0, eta = 1.0;
  double u;
  double hf, hk, hbar, hs, hphi;
  double fplus, fminus, phi, phif, phib, phic, phi2, cf, cb, c2;
  double ef, ebar;

  /* allocate structure to return */
  if (outstruc == NULL) {
    outstruc = (struct estints *) malloc(sizeof(struct estints));
    outstruc->errors = (int *) calloc(nparms, sizeof(int));
    outstruc->hf = (double *) calloc(nparms, sizeof(double));
    outstruc->phi = (double *) calloc(nparms, sizeof(double));
    outstruc->phic = (double *) calloc(nparms, sizeof(double));
    outstruc->phi2 = (double *) calloc(nparms, sizeof(double));
    outstruc->ef = (double *) calloc(nparms, sizeof(double));
    outstruc->nparms = nparms;
  }

  /* evaluate func at the input point */
  u = func(invals);

  /* copy the parameter values for point at which to evaluate the gradient */
  for (i=0; i<nparms; i++) wrk[i] = invals[i];

  for (i=0; i<nparms; i++) {
  FD1:
    hbar = 2.0*(eta+fabs(invals[i]))* sqrt(eps[i]/(omega + fabs(u))) ;
    hk = 10.0 * hbar;
    k = 0;
    fdestimates(i, u, invals, wrk, eps[i], hk,
		 &fplus, &fminus, &phif, &phib, &phic, &phi2, &cf, &cb, &c2,
		 func);
    hs = -1.0;
  FD2:
    if ((cf>cb ? cf : cb) <= 0.1) hs = hk;
    if (0.001 <= c2 && c2 <= 0.1) {
      hphi = hk;
      goto FD5;
    }
    if (0.001 > c2) goto FD4;
  FD3:
    do {
      k++;
      hk *= 10.0;
      fdestimates(i, u, invals, wrk, eps[i], hk,
		  &fplus, &fminus, &phif, &phib, &phic, &phi2, &cf, &cb, &c2,
		  func);
      if (hs<0 && ((cf>cb ? cf : cb) <= 0.1)) hs = hk;
      if (c2 <= 0.1) {
	hphi = hk;
	goto FD5;
      }
    } while (k<K);
    if (k==K) goto FD6;
  FD4:
    do {
      k++;
      hk /= 10.0;
      fdestimates(i, u, invals, wrk, eps[i], hk,
		  &fplus, &fminus, &phif, &phib, &phic, &phi2, &cf, &cb, &c2,
		  func);
      if (c2 > 0.1) {
	hphi = hk * 10.0;
	goto FD5;
      }
      if ((cf>cb ? cf : cb) <= 0.1) hs = hk;
      if (0.001 <= c2 && c2 <= 0.1) {
	hphi = hk;
	goto FD5;
      }
    } while (k<K);
    if (k==K) goto FD6;
  FD5:
    hf = 2.0*sqrt(eps[i]/fabs(phi2));
    wrk[i] = invals[i] + hf;
    fplus = func(wrk);
    phi = (fplus-u)/hf;
    wrk[i] = invals[i] + hphi;
    fplus = func(wrk);
    wrk[i] = invals[i] - hphi;
    fminus = func(wrk);
    wrk[i] = invals[i];
    phic = (fplus-fminus)/(2.0*hphi);

    ef = hf*fabs(phi2)*0.5 + 2.0*eps[i]/hf ;
    ebar = fabs(phi-phic);

    outstruc->hf[i] = hf;
    outstruc->phi[i] = phi;
    outstruc->phic[i] = phic;
    outstruc->phi2[i] = phi2;
    outstruc->ef[i] = ef;
    if ((ef>ebar ? ef : ebar) <= 0.5*fabs(phi)) {
      outstruc->errors[i] = 0;
    }
    else
      outstruc->errors[i] = 1;
    continue;
  FD6:
    if (hs<0) {
      hf = hbar;
      phi = phi2 = ef = 0.0;
      errval = 2;
    }
    else if (hs > 0 && c2 > 0.1) {
      hf = hs;
      wrk[i] = invals[i] + hf;
      fplus = func(wrk);
      wrk[i] = invals[i];
      phi = (fplus-u)/hf;
      phi2 = 0.0;
      ef = 2.0*eps[i]/hf ;
      errval = 3;
    }
    else {
      hf = hk;
      wrk[i] = invals[i] + hf;
      fplus = func(wrk);
      phi = (fplus-u)/hf;
      wrk[i] = invals[i] - hf;
      fminus = func(wrk);
      wrk[i] = invals[i];
      phic = (fplus-fminus)/(2.0*hf);
      ef = hf*fabs(phi2)*0.5 + 2.0*eps[i]/hf ;
      errval = 4;
    }
    outstruc->hf[i] = hf;
    outstruc->phi[i] = phi;
    outstruc->phic[i] = phic;
    outstruc->phi2[i] = phi2;
    outstruc->ef[i] = ef;
    outstruc->errors[i] = errval;
  }
  return outstruc;
}

void fdestimates(int parm, double fvalue, double *invals, double *wrk,
		 double eps, double h,
		 double *fplus, double *fminus,
		 double *phif, double *phib, double *phic, double *phi2,
		 double *cf, double *cb, double *c2,
		 double (*func)(double *))
{
  double ih = 1.0/h;

  wrk[parm] = invals[parm] + h;
  *fplus = func(wrk);
  wrk[parm] = invals[parm] - h;
  *fminus = func(wrk);
  wrk[parm] = invals[parm];
  *phif = (*fplus-fvalue) * ih;
  *phib = (fvalue-*fminus) * ih;
  *phic = (*fplus-*fminus) * 0.5 * ih;
  *phi2 = (*phif-*phib) * ih;
  *cf = 2.0*eps*ih/fabs(*phif);
  *cb = 2.0*eps*ih/fabs(*phib);
  *c2 = 4.0*eps*ih*ih/fabs(*phi2);
}

/* put strict lower triangle of numerical hessian into instruc->hessian */
/* instruc should have been set by algfd */
struct estints *numhessian(struct estints *instruc, double *invals, double *wrk,
			   double (*func)(double *))
{
  static int nparms;
  static double *fplusi = NULL;

  int i,j;
  double hi, hj, ih, jh;
  double fvalue, fplus, phi2;

  if (nparms != instruc->nparms) {
    nparms = instruc->nparms;
    if (fplusi != NULL) free(fplusi);
    fplusi = (double *) malloc(nparms*sizeof(double));
  }

  /* allocate storage for the hessian */
  instruc->hessian = (double *) calloc((nparms*(nparms+1))/2,sizeof(double));

  /* evaluate func at the input point */
  fvalue = func(invals);

  /* copy the parameter values for point at which to evaluate the gradient */
  for (i=0; i<nparms; i++) wrk[i] = invals[i];

  for (i=0; i<nparms; i++) {
    wrk[i] = invals[i] + instruc->hf[i];
    fplusi[i] = func(wrk);
    wrk[i] = invals[i];
  }
  for (i=1; i<nparms; i++) {
    hi = instruc->hf[i];
    ih = 1.0/hi;
    wrk[i] = invals[i] + hi;
    for (j=0; j<i; j++) {
      hj = instruc->hf[j];
      jh = 1.0/hj;
      wrk[j] = invals[j] + hj;
      fplus = func(wrk);
      wrk[j] = invals[j];
      phi2 = (fplus - fplusi[i] - fplusi[j] + fvalue) * ih * jh;
      instruc->hessian[(i*(i-1))/2 + j] = phi2;
    }
    wrk[i] = invals[i];
  }
  return instruc;
}

/* put strict lower triangle of numerical hessian into instruc->hessian */
/*  uses central differences: */
/*   phi2 =
       f(x+hj ej+hi ei) - f(x+hj ej-hi ei) - f(x-hj ej+hi ei) + f(x-hj ej-hi ei) */
/* instruc should have been set by algfd */
struct estints *numhessianc(struct estints *instruc, double *invals, double *wrk,
			   double (*func)(double *))
{
  static int nparms = 0;
  static int nelems = 0;
  static double *fplusi = NULL;
  static double *fminusi = NULL;
  static double *fppi = NULL;
  static double *fpmi = NULL;
  static double *fmmi = NULL;

  int i,j, idx;
  double hi, hj, ih, jh;
  double fvalue, fplus, fminus, phi2;

  if (nparms != instruc->nparms) {
    nparms = instruc->nparms;
    nelems = (nparms*(nparms-1))/2;
    if (fplusi != NULL) free(fplusi);
    fplusi = (double *) malloc(nparms*sizeof(double));
    if (fminusi != NULL) free(fminusi);
    fminusi = (double *) malloc(nparms*sizeof(double));
    if (fppi != NULL) free(fppi);
    fppi = (double *) malloc(nelems*sizeof(double));
    if (fpmi != NULL) free(fpmi);
    fpmi = (double *) malloc(nparms*nparms*sizeof(double));
    if (fmmi != NULL) free(fmmi);
    fmmi = (double *) malloc(nelems*sizeof(double));
  }

  /* allocate storage for the hessian */
  instruc->hessian = (double *) calloc(nelems,sizeof(double));

  /* evaluate func at the input point */
  fvalue = func(invals);

  /* copy the parameter values for point at which to evaluate the gradient */
  for (i=0; i<nparms; i++) wrk[i] = invals[i];

  for (i=0; i<nparms; i++) {
    hi = pow(instruc->hf[i], 2.0/3.0);
    idx = (i*(i-1))/2;
    wrk[i] = invals[i] + 2.0*hi;
    fplusi[i] = func(wrk);
    wrk[i] = invals[i] - 2.0*hi;
    fminusi[i] = func(wrk);
    for (j=0; j<i; j++) {
      hj = pow(instruc->hf[j], 2.0/3.0);
      wrk[i] = invals[i] + hi;
      wrk[j] = invals[j] + hj;
      fppi[idx + j] = func(wrk);
      wrk[i] = invals[i] + hi;
      wrk[j] = invals[j] - hj;
      fpmi[i*nparms + j] = func(wrk);
      wrk[i] = invals[i] - hi;
      wrk[j] = invals[j] + hj;
      fpmi[j*nparms + i] = func(wrk);
      wrk[i] = invals[i] - hi;
      wrk[j] = invals[j] - hj;
      fmmi[idx + j] = func(wrk);
      wrk[j] = invals[j];
    }
    wrk[i] = invals[i];
  }
  for (i=0; i<nparms; i++) {
    hi = pow(instruc->hf[i], 2.0/3.0);
    ih = 1.0/hi;
    idx = (i*(i-1))/2;
    phi2 = (fplusi[i] - 2.0*fvalue + fminusi[i]) * ih * ih * 0.25;
    instruc->phi2[i] = phi2;
    for (j=0; j<i; j++) {
      hj = pow(instruc->hf[j], 2.0/3.0);
      jh = 1.0/hj;
      phi2 = (fppi[idx+j] - fpmi[j*nparms+i] - fpmi[i*nparms+j] + fmmi[idx+j])
	* ih * jh * 0.25;
      instruc->hessian[idx + j] = phi2;
    }
  }
  return instruc;
}

void estoptint(int nparms, int ndiffs, int pflag, double *invals,
		double (*func)(double *))
{
  static double *wrk = NULL;

  int i,j,k, idx;
  int nsteps=1+2*ndiffs, nrows=nparms*nsteps, ncols=ndiffs+1;
  double h, beta, dwrk;
  double **table;
  struct estints *estructure;

  if (wrk==NULL)
    wrk = (double *) malloc(nparms*(ndiffs+1)*sizeof(double));

  h = 0.0000002;
  table = eaccuracy(nparms, ndiffs, h, invals, wrk, func);

  for (i=0; i<nparms*ndiffs; i++) wrk[i] = 0.0;

  for (i=0; i<nparms; i++) {
    for (j=0; j<ndiffs; j++) {
      for (k=1; k<=ndiffs; k++) {
	/* find largest difference of each order for each parm */
	dwrk = fabs(table[j+1][i*nsteps+k]);
	if (wrk[i*ndiffs+j] < dwrk) wrk[i*ndiffs+j] = dwrk;
      }
      beta = sqrt(gamma(1.0+2.0*(1.0+(double)j))/pow(gamma(2.0+(double)j),2.0));
      wrk[i*ndiffs+j] /= beta;
    }
  }
  /* put estimates for highest difference order into epsacc */
  for (i=0; i<nparms; i++) {
    dwrk = wrk[i*ndiffs+ndiffs-1];
    /* make sure epsacc values >= MACHEPS */
    epsacc[i] = (dwrk > MACHEPS ? dwrk : MACHEPS) ;
  }

#ifdef NEVERDEFINED
  printf("accuracy estimates:\n");
  for (i=0; i<nparms; i++) {
    printf("parm = %d\n", i+1);
    for (j=0; j<ndiffs; j++) {
      printf(" %14.7e", wrk[i*ndiffs+j]);
    }
    printf("\n");
  }

  printf("difference table:\n");
  for (i=0; i<nparms; i++) {
    printf("parm = %d\n", i+1);
    for (j=0; j<=ndiffs; j++) {
      for (k=0; k<=ndiffs; k++) {
	printf(" %14.7e", table[j][i*nsteps+k]);
      }
      printf("\n");
    }
  }
#endif

  estructure = algfd(nparms, epsacc, invals, wrk, func);
  if (pflag==1) {
    printf("err   interval          f'                fc'               f''               errorbound\n");
    for (i=0; i<nparms; i++) {
      printf(" %d  ", estructure->errors[i]);
      printf(" %17.10e", estructure->hf[i]);
      printf(" %17.10e", estructure->phi[i]);
      printf(" %17.10e", estructure->phic[i]);
      printf(" %17.10e", estructure->phi2[i]);
      printf(" %17.10e", estructure->ef[i]);
      printf("\n");
    }
  }

  /* put estimates for optimal interval into optint */
  for (i=0; i<nparms; i++) {
    optint[i] = estructure->hf[i];
  }
}

void dohessians(int nparms, int nobs, int ndiffs, double *invals,
		double (*func)(double *), double (*funco)(double *, double *))
{
  static double *wrk = NULL;

  int i,j,k, idx;
  int nsteps=1+2*ndiffs, nrows=nparms*nsteps, ncols=ndiffs+1;
  struct estints *estructure;

  double *opg;

  if (wrk==NULL)
    wrk = (double *) malloc(nparms*(ndiffs+1)*sizeof(double));

  estructure = algfd(nparms, epsacc, invals, wrk, func);

#ifdef NEVERDEFINED
  /* numerical hessian, using forward differences for off-diagonal elements */
  numhessian(estructure, invals, wrk, func);
  printf("numerical hessian, forward differences:\n");
  for (i=0; i<nparms; i++) {
    for (j=0; j<nparms; j++) {
      if (i==j)
	printf(" %19.12e", estructure->phi2[i] / 2.0);
      else if (i>j)
	printf(" %19.12e", estructure->hessian[(i*(i-1))/2 + j] / 2.0);
      else if (j>i)
	printf(" %19.12e", estructure->hessian[(j*(j-1))/2 + i] / 2.0);
    }
    printf("\n");
  }
  fflush(stdout);
#endif NEVERDEFINED

  /* numerical hessian, using central differences */
  numhessianc(estructure, invals, wrk, func);
  printf("numerical hessian, central differences:\n");
  for (i=0; i<nparms; i++) {
    for (j=0; j<nparms; j++) {
      if (i==j)
	printf(" %19.12e", estructure->phi2[i] / 2.0);
      else if (i>j)
	printf(" %19.12e", estructure->hessian[(i*(i-1))/2 + j] / 2.0);
      else if (j>i)
	printf(" %19.12e", estructure->hessian[(j*(j-1))/2 + i] / 2.0);
    }
    printf("\n");
  }
  fflush(stdout);

  /* numerical outer product of gradients, using central differences */
  if (funco != NULL) {
    opg = numopgc(nparms, nobs, invals, opg, wrk, funco);
    printf("numerical outer product of gradients, central differences:\n");
    for (i=0; i<nparms; i++) {
      for (j=0; j<nparms; j++)
	printf(" %19.12e", opg[i*nparms+j] / 2.0);
      printf("\n");
    }
    fflush(stdout);
  }
}
